from ..domain import PetriNet


class FromPetriNet:

    def __init__(self):
        pass

    def execute(self, inputs, outputs, places, transitions) -> PetriNet:
        petri_net = PetriNet()
        petri_net.inputs_matrix = inputs
        petri_net.outputs_matrix = outputs
        petri_net.places = places
        petri_net.transitions = transitions
        petri_net.generate_places_as_dict()
        petri_net.generate_transitions_as_dict()
        petri_net.generate_markings()
        return petri_net



from .petri_net import PetriNet

import numpy as np


class PetriNetEngine:

    def __init__(self, petri_net: PetriNet):
        self.petri_net = petri_net
        pass

    def is_enabled_transition(self, *args):
        markings = self.petri_net.get_markings()
        transition_markings = np.array([0] * len(self.petri_net.places))
        for transition in args:
            transition_markings += self.petri_net.get_inputs_matrix_for_transition(transition)
        result_diff = markings - transition_markings
        return np.min(result_diff) >= 0

    def fire_transitions(self, *args):

        markings = self.petri_net.get_markings()
        if not self.is_enabled_transition(*args):
            raise Exception

        inputs_transitions = np.array([0] * len(self.petri_net.places))
        for transition in args:
            inputs_transitions += self.petri_net.get_inputs_matrix_for_transition(transition)

        outputs_transitions = np.array([0] * len(self.petri_net.places))
        for transition in args:
            outputs_transitions += self.petri_net.get_outputs_matrix_for_transition(transition)

        self.petri_net.markings = markings - inputs_transitions + outputs_transitions
        return self.petri_net.markings



from .petri_net import PetriNet


class PetriNetJson:
    pass


class PetriNetMatrix:
    pass


class PetriNetFactory:

    @staticmethod
    def from_json(json: PetriNetJson) -> PetriNet:
        pass

    @staticmethod
    def from_matrix(json: PetriNetMatrix) -> PetriNet:
        pass

    @staticmethod
    def build() -> PetriNet:
        return PetriNet()



import pandas as pd
import numpy as np
from ..domain.petri_net import PetriNet
from ..domain.place import Place
from ..domain.transition import Transition


class FromEventLogToPetriNet:

    def __init__(self):
        pass

    def execute(self, url_dataframe):
        # event_log_dataframe = pd.read_csv(url_dataframe, sep=';')
        event_log_dataframe, event_log_dataframe_copy = self.load_dataframe(url_dataframe)
        event_log_dataframe, dic_places = self.convert_place_name(event_log_dataframe_copy)
        event_log_dataframe = self.create_transition_id(event_log_dataframe, dic_places.copy())
        event_log_dataframe = self.create_places_id(event_log_dataframe, dic_places)
        event_log_dataframe = self.create_tokens(event_log_dataframe)
        cases = event_log_dataframe.sort_values(['timestamp'], ascending=True).groupby('case_id')
        print('\n')
        petri_nets_variants = []
        for key, item in cases:
            petri_net = PetriNet()
            case = cases.get_group(key)
            for i in range(len(case)):
                place = Place(
                    str(case.iloc[i]['place_id']),
                    case.iloc[i]['place_name'],
                    int(case.iloc[i]['tokens'])
                )
                transition = Transition(
                    str(case.iloc[i]['transition_id']),
                    case.iloc[i]['activity'].replace(' ', '_'),
                )
                petri_net.add_place(place)
                petri_net.add_transition(transition)
                if i + 2 > len(case):
                    break
                # input = {'place': str(case.iloc[i + 1]['place_id']), 'transition': str(case.iloc[i + 1]['transition_id']),
                #          'number': 1}
                # output = {'place': str(case.iloc[i + 1]['place_id']),
                #          'transition': str(case.iloc[i + 1]['transition_id']),
                #          'number': 1}
                petri_net.add_input(str(case.iloc[i]['place_id']), str(case.iloc[i]['transition_id']), 1)
                petri_net.add_output(str(case.iloc[i]['transition_id']), str(case.iloc[i + 1]['place_id']), 1)

            petri_nets_variants.append(petri_net)
        return petri_nets_variants

    def convert_place_name(self, df: pd.DataFrame):
        # returns place name
        activities = df.activity.unique()  # lista of activities
        id_activities = [f'P{x + 1}' for x in range(len(activities))]
        dic_places = dict(zip(activities, id_activities))
        df['place_name'] = df['activity']
        df.replace({"place_name": dic_places}, inplace=True)
        return df, dic_places

    def load_dataframe(self, filename: str, sep=';'):
        log = pd.read_csv(filename, sep=sep)
        copy_log = log.copy()
        return log, copy_log

    def create_transition_id(self, df: pd.DataFrame, dic_transition_id: dict):
        # create transition_id
        df['transition_id'] = df['activity']
        i = 1
        for k, v in dic_transition_id.items():
            dic_transition_id[k] = i
            i = i + 1
        df.replace({"transition_id": dic_transition_id}, inplace=True)
        return df

    def create_places_id(self, df: pd.DataFrame, dic_activities: dict):
        # create places_id
        df['place_id'] = df['place_name']
        id_places = []
        places_name = []
        for k, v in dic_activities.items():
            id_places.append(int(v[1]))
            places_name.append(v)
        dic_places = dict(zip(places_name, id_places))
        df.replace({"place_id": dic_places}, inplace=True)
        return df

    def create_tokens(self, df: pd.DataFrame):
        # create tokens
        df['tokens'] = 0
        df['tokens'] = np.where(df['place_name'] == 'P1', 1, 0)
        return df

    def create_inputs(self, df: pd.DataFrame):
        # create inputs
        df['inputs'] = 0
        size = df.shape[0]
        shortlist = []
        for i in range(size):
            shortlist.append((df['place_id'][i], df['transition_id'][i], df['tokens'][i]))
        df['inputs'] = shortlist
        return df

    def get_cost_total(self, df: pd.DataFrame):
        dic_costos = {}
        df = df.groupby('case_id').sum()
        dic_costos = dict(df['costs'])
        return dic_costos



from ..domain import PetriNet
from ..domain.place import Place
from ..domain.transition import Transition


class FromJsonToPetriNet:

    def __init__(self):
        pass

    def execute(self, petri_net_as_json) -> PetriNet:
        petri_net = PetriNet()
        try:
            for place in petri_net_as_json['places']:
                petri_net.add_place(Place(place['id'], place['name'], place['tokens']))
            for transition in petri_net_as_json['transitions']:
                petri_net.add_transition(Transition(transition['id'], transition['name']))
            for input in petri_net_as_json['inputs']:
                petri_net.add_input(input['place_id'], input['transition_id'], input['number'])
            for output in petri_net_as_json['outputs']:
                petri_net.add_output(output['transition_id'], output['place_id'], output['number'])
            petri_net.generate_inputs_matrix()
            petri_net.generate_outputs_matrix()
        except:
            raise Exception("Error in load data from json")
        return petri_net





import numpy as np

from .place import Place
from .transition import Transition


class PetriNet:

    def __init__(self):
        self.places: list[Place] = []
        self.places_as_dict = {}
        self.markings = []
        self.transitions: list[Transition] = []
        self.transitions_as_dict = {}
        self.inputs_matrix = []
        self.outputs_matrix = []
        self.entries = []
        self.outputs = []

    def add_place(self, place: Place):
        self.places_as_dict[place.place_id] = place
        self.places.append(place)
        self.markings.append(place.tokens)
        return True

    def generate_places_as_dict(self):
        self.places_as_dict = {}
        for place in self.places:
            self.places_as_dict[place.place_id] = place

    def generate_transitions_as_dict(self):
        self.transitions_as_dict = {}
        for transition in self.transitions:
            self.transitions_as_dict[transition.transition_id] = {
                'transition': transition,
                'index': self.transitions_as_dict.__len__()
            }

    def add_transition(self, transition: Transition):
        self.transitions_as_dict[transition.transition_id] = {
            'transition': transition,
            'index': self.transitions_as_dict.__len__()
        }
        self.transitions.append(transition)
        return True

    def add_input(self, place_id, transition_id, number):
        self.entries.append({'place': place_id, 'transition': transition_id, 'number': number})
        return True

    def add_output(self, transition_id, place_id, number):
        self.outputs.append({'transition': transition_id, 'place': place_id, 'number': number})
        return True

    def get_markings(self):
        return np.array(self.markings)

    def generate_markings(self):
        self.markings = []
        for place in self.places:
            self.markings.append(place.tokens)

    def generate_inputs_matrix(self):
        inputs_matrix = []
        for transition in self.transitions_as_dict:
            index = 0
            inputs_of_current_transition = [0] * len(self.places)
            for place in self.places:
                for entry in self.entries:
                    if entry['place'] == place.place_id and entry['transition'] == transition:
                        inputs_of_current_transition[index] = entry['number']
                index += 1
            inputs_matrix.append(inputs_of_current_transition)
        self.inputs_matrix = inputs_matrix

    def generate_outputs_matrix(self):
        outputs_matrix = []
        for transition in self.transitions_as_dict:
            index = 0
            outputs_for_current_transition = [0] * len(self.places)
            for place in self.places:
                for output in self.outputs:
                    if output['place'] == place.place_id and output['transition'] == transition:
                        outputs_for_current_transition[index] = output['number']
                index += 1
            outputs_matrix.append(outputs_for_current_transition)
        self.outputs_matrix = outputs_matrix

    def get_inputs_matrix_for_transition(self, transition: str):
        return self.inputs_matrix[self.transitions_as_dict[transition]['index']]

    def get_outputs_matrix_for_transition(self, transition: str):
        return self.outputs_matrix[self.transitions_as_dict[transition]['index']]
